/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#include "slb.h"

static Hash_table *expr_table;

static size_t
expr_hasher (void const *data, unsigned n_buckets)
{
	struct slb_expression const *exp = data;
	return hash_string(exp->ex_name, n_buckets);
}

/* Compare two strings for equality.  */
static bool
expr_compare(void const *data1, void const *data2)
{
	struct slb_expression const *exp1 = data1;
	struct slb_expression const *exp2 = data2;
	return strcmp(exp1->ex_name, exp2->ex_name) == 0;
}

static void
expr_free(void *data)
{
	struct slb_expression *exp = data;
	/* FIXME: slb_tree_free(exp->ex_tree);*/
	/* FIXME: Free string values */
	free(exp);
}


/* Register an expression. */
int
register_expression(struct slb_expression *expr)
{
	struct slb_expression *s;

	if (!((expr_table
	       || (expr_table = hash_initialize(0, 0,
						expr_hasher,
						expr_compare,
						expr_free)))
	      && (s = hash_insert(expr_table, expr))))
		xalloc_die();

	if (s != expr) {
		grecs_error (&expr->ex_locus, 0,
			     _("redefinition of expression `%s'"),
			     expr->ex_name);
		grecs_error (&s->ex_locus, 0,
			     _("previously defined here"));
		expr_free(expr);
		return 1;
	}
	return 0;
}

struct slb_expression *
find_expression(const char *name)
{
	struct slb_expression ex;
	
	if (!expr_table)
		return NULL;
	ex.ex_name = name;
	return hash_lookup(expr_table, &ex);
}

struct slb_expression *
copy_expression(struct slb_expression *src)
{
	struct slb_expression *dst = xmalloc(sizeof(*dst));
	memcpy(dst, src, sizeof(*dst));
	dst->ex_vartab = symtab_create_default(sizeof(struct slb_varref));
	dst->ex_tree = copy_node(src->ex_tree, dst->ex_vartab);
	dst->state = NULL;
}


static void
nodefixup(struct slb_node *node, struct symtab *vit)
{
	struct slb_varinstance *vi;
	struct slb_node *np;
	
	switch (node->n_type) {
	case node_unknown:
		abort();
		
	case node_variable:
		vi = varinst_lookupz(vit, node->v.n_var->vr_sym.name, NULL);
		if (vi) {
			node->v.n_var->vr_inst = vi;
			vi->vi_flags |= SLV_VIF_USED;
		}
		break;

	case node_uminus:
	case node_not:
		nodefixup(node->v.n_arg[0], vit);
		break;
		
	case node_add:
	case node_sub:
	case node_mul:
	case node_div:
	case node_or:
	case node_and:
	case node_eq:
	case node_ne:
 	case node_lt:
	case node_le:
	case node_gt:
	case node_ge:
		nodefixup(node->v.n_arg[0], vit);
		nodefixup(node->v.n_arg[1], vit);
		break;
		
	case node_cond:
		nodefixup(node->v.n_arg[0], vit);
		if (node->v.n_arg[1])
			nodefixup(node->v.n_arg[1], vit);
		nodefixup(node->v.n_arg[2], vit);
		break;

	case node_func:
		for (np = node->v.n_func.args; np; np = np->n_next)
			nodefixup(np, vit);
		break;

	case node_const:
		break;
	}
}

struct report_undef_closure
{
	grecs_locus_t *loc;
	int rc;
};

static int
report_undefined_refs(void *sym, void *data)
{
	struct slb_varref *ref = sym;
	struct report_undef_closure *clos = data;
	
	if (!ref->vr_inst) {
		grecs_error(clos->loc, 0,
			    _("variable or constant %s not defined"),
			    ref->vr_sym.name);
		clos->rc = 1;
	}
	return 0;
}

static int
mark_used_instances(void *sym, void *data)
{
	struct slb_varinstance *vinst = sym;
	grecs_locus_t *loc = data;

	bind_output_refname(vinst);
	if (!(vinst->vi_flags & SLV_VIF_USED)) {
		grecs_warning(loc, 0,
			      _("variable %s is never used"),
			      vinst->vi_sym.name);
	}
	return 0;
}

int
fixup_expression(struct slb_expression *expr, struct symtab *vit)
{
	int i;
	struct report_undef_closure clos;
	
	nodefixup(expr->ex_tree, vit);
	clos.rc = 0;
	clos.loc = &expr->ex_locus;
	symtab_enumerate(expr->ex_vartab, report_undefined_refs, &clos);
	undo_output_refs();
	symtab_enumerate(vit, mark_used_instances, &expr->ex_locus);
	return clos.rc;
}
